﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using Articles.ScriptingCSharp.ScriptFramework.Util;
using ICSharpCode.AvalonEdit.CodeCompletion;
using ICSharpCode.AvalonEdit.Document;
using ICSharpCode.SharpDevelop.Dom;

namespace Articles.ScriptingCSharp.ScriptFramework.CSharp
{
    public class CSharpScriptFile
    {
        private static Dictionary<string, IProjectContent> m_standardReferences;

        private ObservableCollection<string> m_referencedAssemblies;
        private string m_scriptContent;

        private ProjectContentRegistry m_contentRegistry;
        private IFilterStrategy m_filterStrategy;
        private IProjectContent m_projectContent;
        private TextDocument m_codeDocument;

        /// <summary>
        /// Initializes a new instance of the <see cref="CSharpScriptFile" /> class.
        /// </summary>
        public CSharpScriptFile()
        {
            m_scriptContent = string.Empty;

            m_standardReferences = new Dictionary<string, IProjectContent>();

            m_contentRegistry = new ProjectContentRegistry();
            m_filterStrategy = new NonFilterStrategy();
            m_projectContent = new DefaultProjectContent();

            m_referencedAssemblies = new ObservableCollection<string>();
            m_referencedAssemblies.CollectionChanged += OnReferencedAssembliesCollectionChanged;
            m_referencedAssemblies.Add("mscorlib.dll");
            m_referencedAssemblies.Add("Microsoft.CSharp.dll");
            m_referencedAssemblies.Add("System.dll");
            m_referencedAssemblies.Add("System.Core.dll");
            m_referencedAssemblies.Add("System.Data.dll");
            m_referencedAssemblies.Add("System.Linq.dll");
            m_referencedAssemblies.Add("System.Windows.Forms.dll");

            m_scriptContent =
                "using System;" + Environment.NewLine +
                "using System.Collections.Generic; " + Environment.NewLine +
                "using System.Linq; " + Environment.NewLine +
                "using System.Text; " + Environment.NewLine +
                "using System.Threading.Tasks; " + Environment.NewLine +
                "using System.Windows.Forms; " + Environment.NewLine +
                "" + Environment.NewLine +
                "public class ScriptedClass" + Environment.NewLine +
                "{" + Environment.NewLine +
                "    public static void Execute()" + Environment.NewLine +
                "    {" + Environment.NewLine +
                "        MessageBox.Show(\"Hello World! from script\");" + Environment.NewLine +
                "    }" + Environment.NewLine +
                "}" + Environment.NewLine;
            m_codeDocument = new TextDocument(new StringTextSource(m_scriptContent));
        }

        /// <summary>
        /// Parses the assembly behind the given assembly name and returns a IProjectContent 
        /// object for using it for CodeCompletion feature.
        /// </summary>
        /// <param name="assemblyName">The name of the assembly (e. g. System.Windows.Forms).</param>
        /// <param name="registry">The registry used for loading assembly information.</param>
        private static IProjectContent ParseReferencedAssembly(
            string assemblyName, ProjectContentRegistry registry)
        {
            try
            {
                assemblyName = Path.GetFileNameWithoutExtension(assemblyName);

                //Search within all currently loaded assemblies for assembly information
                foreach (Assembly actAssembly in AppDomain.CurrentDomain.GetAssemblies())
                {
                    if (actAssembly.GetName().Name == assemblyName)
                    {
                        IProjectContent result = registry.GetProjectContentForReference(
                            assemblyName, actAssembly.Location);
                        return result;
                    }
                }

                //Try to load the assembly to get all information from it
                Assembly loadedAssembly = Assembly.LoadWithPartialName(assemblyName);
                if (loadedAssembly != null)
                {
                    return registry.GetProjectContentForReference(assemblyName, loadedAssembly.Location);
                }
            }
            catch (Exception)
            {
                //Don't handle exception here, just return null for nothing found
                return null;
            }

            return null;
        }

        /// <summary>
        /// Called when the collection of referenced assemblies has changed.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">
        /// The <see cref="NotifyCollectionChangedEventArgs" /> 
        /// instance containing the event data.
        /// </param>
        private async void OnReferencedAssembliesCollectionChanged(
            object sender, NotifyCollectionChangedEventArgs e)
        {
            if ((e.NewItems != null) && (e.NewItems.Count > 0))
            {
                foreach (string actNewItem in e.NewItems)
                {
                    if (!m_standardReferences.ContainsKey(actNewItem))
                    {
                        string asyncLoadedContentName = actNewItem;

                        //Load IProjectContent object asynchron
                        IProjectContent asyncLoadedContent = await Task.Run<IProjectContent>(() =>
                        {
                            return ParseReferencedAssembly(
                                asyncLoadedContentName, m_contentRegistry);
                        });

                        //Append IProjectContent to main IProjectContent object
                        if (asyncLoadedContent != null)
                        {
                            m_standardReferences[asyncLoadedContentName] = asyncLoadedContent;
                            m_projectContent.AddReferencedContent(asyncLoadedContent);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets the referenced assemblies.
        /// </summary>
        /// <value>
        /// The referenced assemblies.
        /// </value>
        public ObservableCollection<string> ReferencedAssemblies
        {
            get { return m_referencedAssemblies; }
        }

        /// <summary>
        /// Gets or sets the content of the script.
        /// </summary>
        public string ScriptContent
        {
            get { return m_codeDocument.Text; }
        }

        /// <summary>
        /// Gets the current filter strategy for intellisense.
        /// </summary>
        public IFilterStrategy FilterStrategy
        {
            get { return m_filterStrategy; }
        }

        /// <summary>
        /// Gets the IProjectContent object describing all code items
        /// that may be accessed through intellisense.
        /// </summary>
        public IProjectContent ProjectContent
        {
            get { return m_projectContent; }
        }

        /// <summary>
        /// Gets the TextDocument object containing the source code.
        /// </summary>
        public TextDocument CodeDocument
        {
            get { return m_codeDocument; }
        }
    }
}
